import numpy as np
from .. import pixel_formats
from . import Reader, Writer
from . import YUVFrame
from . import colorspaces


def imread(file, width, height, pixel_format, index=0):
    """
    Read the yuv frame from the given file.

    :param file: str, Path, file handle
    :param width: frame width
    :param height: frame height
    :param pixel_format: ffmpeg pixel format specifier
    :param index: frame index (default: 0)
    :return: yuv frame
    """
    yuv_format = pixel_formats[pixel_format](width, height)
    reader = Reader(file, yuv_format)
    return reader.read(index, 1)[0]


def mimread(file, width, height, pixel_format, index=0, count=None):
    """
    Read the yuv frames from the given file.

    :param file: str, Path, file handle
    :param width: frame width
    :param height: frame height
    :param pixel_format: ffmpeg pixel format specifier
    :param index: first frame index (default: 0)
    :param count: frame count (read all if None)
    :return: list of yuv frames
    """
    yuv_format = pixel_formats[pixel_format](width, height)
    reader = Reader(file, yuv_format)
    return reader.read(index, count)


def imwrite(file, yuv_frame):
    """
    Write the yuv frame to the given file.

    :param file: str, Path, file handle
    :param yuv_frame: list of yuv frames
    """
    yuv_format = pixel_formats[yuv_frame.pixel_format](yuv_frame.y.shape[1], yuv_frame.y.shape[0])
    writer = Writer(file, yuv_format)
    writer.write(yuv_frame)


def mimwrite(file, yuv_frames):
    """
    Write the yuv frames to the given file.

    :param file: str, Path, file handle
    :param yuv_frames: list of yuv frames
    """
    yuv_format = pixel_formats[yuv_frames[0].pixel_format](yuv_frames[0].y.shape[1], yuv_frames[0].y.shape[0])
    writer = Writer(file, yuv_format)
    writer.write(yuv_frames)


def get_reader(file, width, height, pixel_format):
    """
    Get a reader for the given file.

    :param file: str, Path, file handle
    :param width: frame width
    :param height: frame height
    :param pixel_format: ffmpeg pixel format specifier
    :return: reader
    """
    yuv_format = pixel_formats[pixel_format](width, height)
    reader = Reader(file, yuv_format)
    return reader


def get_writer(file, width, height, pixel_format):
    """
    Get a writer for the given file.

    :param file: str, Path, file handle
    :param width: frame width
    :param height: frame height
    :param pixel_format: ffmpeg pixel format specifier
    :return: writer
    """
    yuv_format = pixel_formats[pixel_format](width, height)
    writer = Writer(file, yuv_format)
    return writer


def frame(yuv, pixel_format):
    """
    Initialize a new yuv frame from the given y, u, v components.

    :param yuv: yuv components tuple
    :param pixel_format: ffmpeg pixel format specifier
    :return: yuv frame
    """
    y, u, v = yuv
    yuv_format = pixel_formats[pixel_format](y.shape[1], y.shape[0])
    sub_w, sub_h = yuv_format.chroma_subsampling()
    if sub_w == 0 and sub_h == 0:
        if u is not None or v is not None:
            raise RuntimeError("Chroma components must be 'None' for pixel format '{}'".format(pixel_format))
        expected_chroma_shape = (0, 0)
        u = np.empty(expected_chroma_shape, dtype=np.void)
        v = np.empty(expected_chroma_shape, dtype=np.void)
    else:
        expected_chroma_shape = (y.shape[0] / sub_h, y.shape[1] / sub_w)
    if u.shape != expected_chroma_shape or v.shape != expected_chroma_shape:
        raise RuntimeError("Invalid chroma shape for pixel format '{}'".format(pixel_format))
    yuv_frame = YUVFrame(y, u, v, yuv_format)
    return yuv_frame


def empty(width, height, pixel_format):
    """
    Initialize a new yuv frame where all components are empty.

    :param width: frame width
    :param height: frame height
    :param pixel_format: ffmpeg pixel format specifier
    :return: yuv frame
    """
    yuv_format = pixel_formats[pixel_format](width, height)
    y, u, v = yuv_format.unpack(np.empty(1, dtype=yuv_format.dtype))
    return YUVFrame(y[0],
                    u[0] if u is not None else None,
                    v[0] if v is not None else None,
                    yuv_format)


def zeros(width, height, pixel_format):
    """
    Initialize a new yuv frame where all components are set to 0.

    :param width: frame width
    :param height: frame height
    :param pixel_format: ffmpeg pixel format specifier
    :return: yuv frame
    """
    yuv_format = pixel_formats[pixel_format](width, height)
    y, u, v = yuv_format.unpack(np.zeros(1, dtype=yuv_format.dtype))
    return YUVFrame(y[0],
                    u[0] if u is not None else None,
                    v[0] if v is not None else None,
                    yuv_format)


def ones(width, height, pixel_format):
    """
    Initialize a new yuv frame where all components are set to 1.

    :param width: frame width
    :param height: frame height
    :param pixel_format: ffmpeg pixel format specifier
    :return: yuv frame
    """
    yuv_format = pixel_formats[pixel_format](width, height)
    y, u, v = yuv_format.unpack(np.ones(1, dtype=yuv_format.dtype))
    return YUVFrame(y[0],
                    u[0] if u is not None else None,
                    v[0] if v is not None else None,
                    yuv_format)


def to_rgb(yuv, specification='bt709', value_range='limited'):
    """
    Convert yuv data to rgb.

    :param yuv: yuv frame
    :param specification: specification identifier (default: 'bt709')
    :param value_range: yuv value range (default: 'limited')
    :return: rgb data
    """
    return colorspaces[specification, value_range].to_rgb(*yuv.split(), yuv.yuv_format)


def from_rgb(rgb, pixel_format, specification='bt709', value_range='limited'):
    """
    Initialize a new yuv frame from rgb data.

    :param rgb: rgb data
    :param pixel_format: ffmpeg pixel format specifier
    :param specification: specification identifier (default: 'bt709')
    :param value_range: yuv value range (default: 'limited')
    :return: yuv frame
    """
    yuv_format = pixel_formats[pixel_format](rgb.shape[1], rgb.shape[0])
    y, u, v = colorspaces[specification, value_range].from_rgb(rgb, yuv_format)
    return YUVFrame(y, u, v, yuv_format)



from abc import ABC, abstractmethod
from typing import Dict, Tuple, Optional
import numpy as np


class Format(ABC):

    def __init__(self, width, height):
        self._width = width
        self._height = height

    @classmethod
    def io_info(cls) -> str:
        """IO format info as
        '(component1 -> (height1, width1), component2 -> (height2, width2), ...)'"""
        sub_w, sub_h = cls.chroma_subsampling()
        if sub_w == 0 and sub_h == 0:
            return "y -> (height, width)"
        else:
            return "y -> (height, width), " \
                   "u/Cb -> (height / {1}, width / {0}), " \
                   "v/Cr -> (height / {1}, width / {0})".format(sub_w, sub_h)

    @staticmethod
    @abstractmethod
    def chroma_subsampling():
        """Return chroma subsampling as '(factor width, factor height)'."""
        pass

    @staticmethod
    @abstractmethod
    def bitdepth():
        """Return bitdepth of this format."""
        pass

    @staticmethod
    @abstractmethod
    def identifier():
        """Return the unique identifier for this format."""
        pass

    @property
    @abstractmethod
    def dtype(self) -> np.dtype:
        """Numpy dtype describing exactly one frame."""
        pass

    @abstractmethod
    def unpack(self, data: np.ndarray) -> Tuple[np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]:
        """Unpack the data described by dtype into raw components."""
        pass

    @abstractmethod
    def pack(self, raw: Tuple[np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]) -> np.ndarray:
        """Pack the raw components into data described by dtype."""
        pass


class FormatManager:

    def __init__(self):
        self._formats: Dict[str, Format] = {}

    def register(self, format_cls: Format, overwrite: bool = False):
        pix_fmt = format_cls.identifier()
        if pix_fmt in self._formats and not overwrite:
            raise KeyError("Another format with identifier '{}' is registered already.".format(pix_fmt))
        self._formats[pix_fmt] = format_cls

    def __iter__(self):
        return iter(self._formats)

    def __len__(self):
        return len(self._formats)

    def __getitem__(self, key):
        return self._formats[key]

    def __setitem__(self, key, value):
        self.register(value, key)

    def __contains__(self, item):
        return item in self._formats

    def __str__(self):
        return ", ".join(list(self._formats.keys()))

    def io_description(self, pix_fmt: str):
        return self[pix_fmt].io_info()



from abc import ABC
import numpy as np
from .. import pixel_formats
from ..core import Format


class _YUV420PBase(Format, ABC):
    """Base for all yuv420 formats."""
    @staticmethod
    def chroma_subsampling():
        return 2, 2

    def unpack(self, data):
        y = data['y']
        u = data['u']
        v = data['v']
        return y, u, v

    def pack(self, yuv):
        data = np.empty(yuv[0].shape[0], dtype=self.dtype)
        data['y'][:] = yuv[0]
        data['u'][:] = yuv[1]
        data['v'][:] = yuv[2]
        return data


class YUV420P(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u1', (self._height, self._width)),
            ('u', '<u1', (self._height // 2, self._width // 2)),
            ('v', '<u1', (self._height // 2, self._width // 2))
        ])


class YUV420P10LE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p10le"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '<u2', (self._height // 2, self._width // 2)),
            ('v', '<u2', (self._height // 2, self._width // 2))
        ])


class YUV420P10BE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p10be"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '>u2', (self._height // 2, self._width // 2)),
            ('v', '>u2', (self._height // 2, self._width // 2))
        ])


class YUV420P16LE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p16le"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '<u2', (self._height // 2, self._width // 2)),
            ('v', '<u2', (self._height // 2, self._width // 2))
        ])


class YUV420P16BE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p16be"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '>u2', (self._height // 2, self._width // 2)),
            ('v', '>u2', (self._height // 2, self._width // 2))
        ])


class YUV420P9LE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p9le"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '<u2', (self._height // 2, self._width // 2)),
            ('v', '<u2', (self._height // 2, self._width // 2))
        ])


class YUV420P9BE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p9be"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '>u2', (self._height // 2, self._width // 2)),
            ('v', '>u2', (self._height // 2, self._width // 2))
        ])

class YUV420P12LE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p12le"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '<u2', (self._height // 2, self._width // 2)),
            ('v', '<u2', (self._height // 2, self._width // 2))
        ])


class YUV420P12BE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p12be"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '>u2', (self._height // 2, self._width // 2)),
            ('v', '>u2', (self._height // 2, self._width // 2))
        ])


class YUV420P14LE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p14le"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '<u2', (self._height // 2, self._width // 2)),
            ('v', '<u2', (self._height // 2, self._width // 2))
        ])


class YUV420P14BE(_YUV420PBase):
    @staticmethod
    def identifier():
        return "yuv420p14be"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '>u2', (self._height // 2, self._width // 2)),
            ('v', '>u2', (self._height // 2, self._width // 2))
        ])


pixel_formats.register(YUV420P)
pixel_formats.register(YUV420P10LE)
pixel_formats.register(YUV420P10BE)
pixel_formats.register(YUV420P16LE)
pixel_formats.register(YUV420P16BE)
pixel_formats.register(YUV420P9LE)
pixel_formats.register(YUV420P9BE)
pixel_formats.register(YUV420P12LE)
pixel_formats.register(YUV420P12BE)
pixel_formats.register(YUV420P14LE)
pixel_formats.register(YUV420P14BE)



from abc import ABC
import numpy as np
from .. import pixel_formats
from ..core import Format


class _YUV422IBase(Format, ABC):
    """Base for all interleaved yuv422 formats."""
    @staticmethod
    def chroma_subsampling():
        return 2, 1

    def unpack(self, data):
        y = np.stack((data['frame']['y0'],
                      data['frame']['y1']), 3).reshape((-1, self._height, self._width))
        u = data['frame']['u0']
        v = data['frame']['v0']
        return y, u, v

    def pack(self, yuv):
        y, u, v = yuv
        data = np.empty(y.shape[0], dtype=self.dtype)
        y = y.reshape((-1, self._height, self._width // 2, 2))
        y0, y1 = y[:, :, :, 0], y[:, :, :, 1]
        data['frame']['y0'][:] = y0
        data['frame']['u0'][:] = u
        data['frame']['y1'][:] = y1
        data['frame']['v0'][:] = v
        return data


class YUV422I(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u1"),
              ('u0', "<u1"),
              ('y1', "<u1"),
              ('v0', "<u1")],
             (self._height, self._width // 2))
        ])


class YUV422I10LE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i10le"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u2"),
              ('u0', "<u2"),
              ('y1', "<u2"),
              ('v0', "<u2")],
             (self._height, self._width // 2))
        ])


class YUV422I10BE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i10be"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', ">u2"),
              ('u0', ">u2"),
              ('y1', ">u2"),
              ('v0', ">u2")],
             (self._height, self._width // 2))
        ])
    

class YUV422I16LE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i16le"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u2"),
              ('u0', "<u2"),
              ('y1', "<u2"),
              ('v0', "<u2")],
             (self._height, self._width // 2))
        ])


class YUV422I16BE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i16be"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', ">u2"),
              ('u0', ">u2"),
              ('y1', ">u2"),
              ('v0', ">u2")],
             (self._height, self._width // 2))
        ])


class YUV422I9LE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i9le"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u2"),
              ('u0', "<u2"),
              ('y1', "<u2"),
              ('v0', "<u2")],
             (self._height, self._width // 2))
        ])


class YUV422I9BE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i9be"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', ">u2"),
              ('u0', ">u2"),
              ('y1', ">u2"),
              ('v0', ">u2")],
             (self._height, self._width // 2))
        ])


class YUV422I12LE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i12le"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u2"),
              ('u0', "<u2"),
              ('y1', "<u2"),
              ('v0', "<u2")],
             (self._height, self._width // 2))
        ])


class YUV422I12BE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i12be"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', ">u2"),
              ('u0', ">u2"),
              ('y1', ">u2"),
              ('v0', ">u2")],
             (self._height, self._width // 2))
        ])


class YUV422I14LE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i14le"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', "<u2"),
              ('u0', "<u2"),
              ('y1', "<u2"),
              ('v0', "<u2")],
             (self._height, self._width // 2))
        ])


class YUV422I14BE(_YUV422IBase):
    @staticmethod
    def identifier():
        return "yuv422i14be"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y0', ">u2"),
              ('u0', ">u2"),
              ('y1', ">u2"),
              ('v0', ">u2")],
             (self._height, self._width // 2))
        ])


pixel_formats.register(YUV422I)
pixel_formats.register(YUV422I10LE)
pixel_formats.register(YUV422I10BE)
pixel_formats.register(YUV422I16LE)
pixel_formats.register(YUV422I16BE)
pixel_formats.register(YUV422I9LE)
pixel_formats.register(YUV422I9BE)
pixel_formats.register(YUV422I12LE)
pixel_formats.register(YUV422I12BE)
pixel_formats.register(YUV422I14LE)
pixel_formats.register(YUV422I14BE)



from abc import ABC
import numpy as np
from .. import pixel_formats
from ..core import Format


class _YUV444IBase(Format, ABC):
    """Base for all interleaved yuv444 formats."""
    @staticmethod
    def chroma_subsampling():
        return 1, 1

    def unpack(self, data):
        y = data['frame']['y']
        u = data['frame']['u']
        v = data['frame']['v']
        return y, u, v

    def pack(self, yuv):
        data = np.empty(yuv[0].shape[0], dtype=self.dtype)
        data['frame']['y'][:] = yuv[0]
        data['frame']['u'][:] = yuv[1]
        data['frame']['v'][:] = yuv[2]
        return data


class YUV444I(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', "<u1"),
              ('u', "<u1"),
              ('v', "<u1")],
             (self._height, self._width))
        ])


class YUV444I10LE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i10le"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', "<u2"),
              ('u', "<u2"),
              ('v', "<u2")],
             (self._height, self._width))
        ])


class YUV444I10BE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i10be"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', ">u2"),
              ('u', ">u2"),
              ('v', ">u2")],
             (self._height, self._width))
        ])


class YUV444I16LE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i16le"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', "<u2"),
              ('u', "<u2"),
              ('v', "<u2")],
             (self._height, self._width))
        ])


class YUV444I16BE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i16be"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', ">u2"),
              ('u', ">u2"),
              ('v', ">u2")],
             (self._height, self._width))
        ])


class YUV444I9LE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i9le"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', "<u2"),
              ('u', "<u2"),
              ('v', "<u2")],
             (self._height, self._width))
        ])


class YUV444I9BE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i9be"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', ">u2"),
              ('u', ">u2"),
              ('v', ">u2")],
             (self._height, self._width))
        ])


class YUV444I12LE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i12le"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', "<u2"),
              ('u', "<u2"),
              ('v', "<u2")],
             (self._height, self._width))
        ])


class YUV444I12BE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i12be"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', ">u2"),
              ('u', ">u2"),
              ('v', ">u2")],
             (self._height, self._width))
        ])


class YUV444I14LE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i14le"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', "<u2"),
              ('u', "<u2"),
              ('v', "<u2")],
             (self._height, self._width))
        ])


class YUV444I14BE(_YUV444IBase):
    @staticmethod
    def identifier():
        return "yuv444i14be"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('frame',
             [('y', ">u2"),
              ('u', ">u2"),
              ('v', ">u2")],
             (self._height, self._width))
        ])


pixel_formats.register(YUV444I)
pixel_formats.register(YUV444I10LE)
pixel_formats.register(YUV444I10BE)
pixel_formats.register(YUV444I16LE)
pixel_formats.register(YUV444I16BE)
pixel_formats.register(YUV444I9LE)
pixel_formats.register(YUV444I9BE)
pixel_formats.register(YUV444I12LE)
pixel_formats.register(YUV444I12BE)
pixel_formats.register(YUV444I14LE)
pixel_formats.register(YUV444I14BE)



from abc import ABC
import numpy as np
from .. import pixel_formats
from ..core import Format


class _GrayBase(Format, ABC):
    """Base for all gray/yuv400 formats."""
    @staticmethod
    def chroma_subsampling():
        return 0, 0

    def unpack(self, data):
        y = data['y']
        return y, None, None

    def pack(self, yuv):
        data = np.empty(yuv[0].shape[0], dtype=self.dtype)
        data['y'][:] = yuv[0]
        return data


class Gray(_GrayBase):
    @staticmethod
    def identifier():
        return "gray"

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u1', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray10LE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray10le"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray10BE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray10be"

    @staticmethod
    def bitdepth():
        return 10

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray16LE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray16le"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray16BE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray16be"

    @staticmethod
    def bitdepth():
        return 16

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray9LE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray9le"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray9BE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray9be"

    @staticmethod
    def bitdepth():
        return 9

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray12LE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray12le"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray12BE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray12be"

    @staticmethod
    def bitdepth():
        return 12

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray14LE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray14le"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


class Gray14BE(_GrayBase):
    @staticmethod
    def identifier():
        return "gray14be"

    @staticmethod
    def bitdepth():
        return 14

    @property
    def dtype(self):
        return np.dtype([
            ('y', '>u2', (self._height, self._width)),
            ('u', '|V0'),
            ('v', '|V0')
        ])


pixel_formats.register(Gray)
pixel_formats.register(Gray10LE)
pixel_formats.register(Gray10BE)
pixel_formats.register(Gray16LE)
pixel_formats.register(Gray16BE)
pixel_formats.register(Gray9LE)
pixel_formats.register(Gray9BE)
pixel_formats.register(Gray12LE)
pixel_formats.register(Gray12BE)
pixel_formats.register(Gray14LE)
pixel_formats.register(Gray14BE)



import numpy as np
from .. import pixel_formats
from ..core import Format


class NV12(Format):
    """NV12 A plane of 8bit Y samples follows by an interleaved plane of 8bit U/V yuv 420 format."""
    @staticmethod
    def identifier():
        return "nv12"

    @staticmethod
    def chroma_subsampling():
        return 2, 2

    @staticmethod
    def bitdepth():
        return 8

    @property
    def dtype(self):
        return np.dtype([
            ('y', '<u1', (self._height, self._width)),
            ('uv', '<u1', (self._height // 2, self._width))
        ])

    def unpack(self, data):
        y = data['y']
        u = data['uv'][:, :, ::2]
        v = data['uv'][:, :, 1::2]
        return y, u, v

    def pack(self, yuv):
        data = np.empty(yuv[0].shape[0], dtype=self.dtype)
        data['y'][:] = yuv[0]
        data['uv'][:, :, ::2] = yuv[1]
        data['uv'][:, :, 1::2] = yuv[2]
        return data


pixel_formats.register(NV12)



from typing import Union, List
from io import IOBase
from pathlib import Path
import numpy as np
from . import YUVFrame
from . import Format


class Writer:

    def __init__(self, file, format: Format):
        if isinstance(file, IOBase):
            self._close = False
            self._file = file
        else:
            self._close = True
            self._file = open(Path(file).expanduser().resolve(), 'wb')
        self._format = format

    def __del__(self):
        if self._close:
            self._file.close()

    def write(self, yuv_frames: Union[List[YUVFrame], YUVFrame]):
        if isinstance(yuv_frames, YUVFrame):
            yuv_frames = [yuv_frames]
        frame_count = len(yuv_frames)
        if frame_count == 0:
            return

        if self._format.chroma_subsampling() != (0, 0):
            y = np.empty((frame_count, yuv_frames[0].y.shape[0], yuv_frames[0].y.shape[1]),
                         dtype=yuv_frames[0].y.dtype)
            u = np.empty((frame_count, yuv_frames[0].u.shape[0], yuv_frames[0].u.shape[1]),
                         dtype=yuv_frames[0].u.dtype)
            v = np.empty((frame_count, yuv_frames[0].v.shape[0], yuv_frames[0].v.shape[1]),
                         dtype=yuv_frames[0].v.dtype)
            for i, yuv_frame in enumerate(yuv_frames):
                y[i] = yuv_frame[0]
                u[i] = yuv_frame[1]
                v[i] = yuv_frame[2]
            data = self._format.pack((y, u, v))
        else:
            y = np.empty((frame_count, yuv_frames[0].y.shape[0], yuv_frames[0].y.shape[1]),
                         dtype=yuv_frames[0].y.dtype)
            for i, yuv_frame in enumerate(yuv_frames):
                y[i] = yuv_frame[0]
            data = self._format.pack((y, None, None))

        self._file.write(data.data)



from .gray import (Gray,
                   Gray10LE, Gray10BE,
                   Gray16LE, Gray16BE,
                   Gray9LE, Gray9BE,
                   Gray12LE, Gray12BE,
                   Gray14LE, Gray14BE)
from .nv12 import NV12
from .v210 import V210
from .yuv420p import (YUV420P,
                      YUV420P10LE, YUV420P10BE,
                      YUV420P16LE, YUV420P16BE,
                      YUV420P9LE, YUV420P9BE,
                      YUV420P12LE, YUV420P12BE,
                      YUV420P14LE, YUV420P14BE)
from .yuv420i import (YUV420I,
                      YUV420I10LE, YUV420I10BE,
                      YUV420I16LE, YUV420I16BE,
                      YUV420I9LE, YUV420I9BE,
                      YUV420I12LE, YUV420I12BE,
                      YUV420I14LE, YUV420I14BE)
from .yuv422p import (YUV422P,
                      YUV422P10LE, YUV422P10BE,
                      YUV422P16LE, YUV422P16BE,
                      YUV422P9LE, YUV422P9BE,
                      YUV422P12LE, YUV422P12BE,
                      YUV422P14LE, YUV422P14BE)
from .yuv422i import (YUV422I,
                      YUV422I10LE, YUV422I10BE,
                      YUV422I16LE, YUV422I16BE,
                      YUV422I9LE, YUV422I9BE,
                      YUV422I12LE, YUV422I12BE,
                      YUV422I14LE, YUV422I14BE)
from .yuv444p import (YUV444P,
                      YUV444P10LE, YUV444P10BE,
                      YUV444P16LE, YUV444P16BE,
                      YUV444P9LE, YUV444P9BE,
                      YUV444P12LE, YUV444P12BE,
                      YUV444P14LE, YUV444P14BE)
from .yuv444i import (YUV444I,
                      YUV444I10LE, YUV444I10BE,
                      YUV444I16LE, YUV444I16BE,
                      YUV444I9LE, YUV444I9BE,
                      YUV444I12LE, YUV444I12BE,
                      YUV444I14LE, YUV444I14BE)
from .yuyv422 import (YUYV422, UYVY422, YVYU422)



from typing import Union, Tuple, Optional
import numpy as np
from . import Format
from . import colorspaces


class YUVFrame:
    """YUVFrame provides convenient data access to a single yuv/ycbcr frame."""

    def __init__(self, y: np.ndarray,
                 u: Optional[np.ndarray],
                 v: Optional[np.ndarray],
                 yuv_format: Format):
        self._y = y
        self._u = u
        self._v = v
        self._yuv_format = yuv_format

    @property
    def pixel_format(self):
        return self._yuv_format.identifier()

    @property
    def yuv_format(self):
        return self._yuv_format

    @property
    def resolution(self):
        return self._y.shape[1], self._y.shape[0]

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value: np.ndarray):
        self._y = value

    @property
    def u(self):
        return self._u

    @u.setter
    def u(self, value: Optional[np.ndarray]):
        self._u = value

    @property
    def v(self):
        return self._v

    @v.setter
    def v(self, value: Optional[np.ndarray]):
        self._v = value

    @property
    def cb(self):
        return self._u

    @cb.setter
    def cb(self, value: np.ndarray):
        self._u = value

    @property
    def cr(self):
        return self._v

    @cr.setter
    def cr(self, value: np.ndarray):
        self._v = value

    def __getitem__(self, key: Union[str, int]):
        if isinstance(key, str):
            if key == 'y':
                key = 0
            elif key == 'u' or key == 'cb':
                key = 1
            elif key == 'v' or key == 'cr':
                key = 2

        return [self._y, self._u, self._v][key]

    def __setitem__(self, key: Union[str, int], value: np.ndarray):
        if isinstance(key, str):
            if key == 'y':
                key = 0
            elif key == 'u' or key == 'cb':
                key = 1
            elif key == 'v' or key == 'cr':
                key = 2

            [self._y, self._u, self._v][key] = value

    def set(self, yuv: Tuple[np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]):
        self._y = yuv[0]
        self._u = yuv[1]
        self._v = yuv[2]

    def split(self):
        return self._y, self._u, self._v



from typing import List, Tuple
import numpy as np
from . import Format


class Colorspace:
    def __init__(self, coefficients: List[float],
                 y_baserange: Tuple[int, int],
                 cbcr_baserange: Tuple[int, int],
                 video_basemargin: int):
        self._coefficients = coefficients
        self._y_baserange = y_baserange
        self._cbcr_baserange = cbcr_baserange
        self._video_basemargin = video_basemargin
        self._precompute_16bit_conversion_coefficients()

    def _precompute_16bit_conversion_coefficients(self):
        a, b, c, d, e = self._coefficients
        y_baselength = self._y_baserange[1] - self._y_baserange[0]
        cbcr_baselength = self._cbcr_baserange[1] - self._cbcr_baserange[0]

        y_scale_torgb_16bit = int(np.round((255 / y_baselength) * (2 ** 16)))
        d_torgb_16bit = int(np.round(d * (255 / cbcr_baselength) * (2 ** 16)))
        e_torgb_16bit = int(np.round(e * (255 / cbcr_baselength) * (2 ** 16)))
        ae_b_torgb_16bit = int(np.round((a * e / b) * (255 / cbcr_baselength) * (2 ** 16)))
        cd_b_torgb_16bit = int(np.round((c * d / b) * (255 / cbcr_baselength) * (2 ** 16)))
        self._to_rgb_coefficients = [y_scale_torgb_16bit,
                                     d_torgb_16bit, e_torgb_16bit,
                                     ae_b_torgb_16bit, cd_b_torgb_16bit]

        a_fromrgb_16bit = int(np.round(a * (2 ** 16)))
        b_fromrgb_16bit = int(np.round(b * (2 ** 16)))
        c_fromrgb_16bit = int(np.round(c * (2 ** 16)))
        inv_d_fromrgb_16bit = int(np.round((1 / d) * (2 ** 16)))
        inv_e_fromrgb_16bit = int(np.round((1 / e) * (2 ** 16)))
        y_scale_fromrgb_16bit = int(np.round((y_baselength / 255) * (2 ** 16)))
        cbcr_scale_fromrgb_16bit = int(np.round((cbcr_baselength / 255) * (2 ** 16)))
        self._from_rgb_coefficients = [a_fromrgb_16bit, b_fromrgb_16bit, c_fromrgb_16bit,
                                       inv_d_fromrgb_16bit, inv_e_fromrgb_16bit,
                                       y_scale_fromrgb_16bit, cbcr_scale_fromrgb_16bit]

    def to_rgb(self, y, u, v, yuv_format: Format) -> np.ndarray:
        if yuv_format.chroma_subsampling()[0] != 1 or yuv_format.chroma_subsampling()[1] != 1:
            raise ValueError("Color conversion is only possible for 444 yuv (ycbcr) formats (no chroma subsampling). "
                             f"'{yuv_format.identifier()}' entails chroma subsampling.")
        bitdepth = yuv_format.bitdepth()
        max_value = 2 ** bitdepth - 1
        dtype = y.dtype

        y = y.astype(np.int64)
        u = u.astype(np.int64)
        v = v.astype(np.int64)

        # Color conversion
        y_offset = self._y_baserange[0] << (bitdepth - 8)
        chroma_center = 128 << (bitdepth - 8)

        y = (y - y_offset)
        u = (u - chroma_center)
        v = (v - chroma_center)

        (y_scale_torgb_16bit,
         d_torgb_16bit, e_torgb_16bit,
         ae_b_torgb_16bit, cd_b_torgb_16bit) = self._to_rgb_coefficients
        y_16bit = y * y_scale_torgb_16bit
        r = (y_16bit + e_torgb_16bit * v) >> 16
        g = (y_16bit - ae_b_torgb_16bit * v - cd_b_torgb_16bit * u) >> 16
        b = (y_16bit + d_torgb_16bit * u) >> 16

        r = np.clip(r, 0, max_value).astype(dtype)
        g = np.clip(g, 0, max_value).astype(dtype)
        b = np.clip(b, 0, max_value).astype(dtype)

        return np.stack((r, g, b), axis=-1)

    def from_rgb(self, rgb_frame: np.ndarray, yuv_format: Format) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if yuv_format.chroma_subsampling()[0] != 1 or yuv_format.chroma_subsampling()[1] != 1:
            raise ValueError("Color conversion is only possible for 444 yuv (ycbcr) formats (no chroma subsampling). "
                             f"'{yuv_format.identifier()}' entails chroma subsampling.")
        bitdepth = yuv_format.bitdepth()
        rgb = rgb_frame.astype(np.int64)

        (a_fromrgb_16bit, b_fromrgb_16bit, c_fromrgb_16bit,
         inv_d_fromrgb_16bit, inv_e_fromrgb_16bit,
         y_scale_fromrgb_16bit, cbcr_scale_fromrgb_16bit) = self._from_rgb_coefficients

        y = a_fromrgb_16bit * rgb[..., 0] + b_fromrgb_16bit * rgb[..., 1] + c_fromrgb_16bit * rgb[..., 2]
        u = (((rgb[..., 2] << 16) - y) * inv_d_fromrgb_16bit) >> 16
        v = (((rgb[..., 0] << 16) - y) * inv_e_fromrgb_16bit) >> 16

        bitdepth_shift = bitdepth - 8
        y_low = self._y_baserange[0] << bitdepth_shift
        chroma_center = 128 << bitdepth_shift

        y = ((y * y_scale_fromrgb_16bit) >> 32) + y_low
        u = ((u * cbcr_scale_fromrgb_16bit) >> 32) + chroma_center
        v = ((v * cbcr_scale_fromrgb_16bit) >> 32) + chroma_center

        clip_low = self._video_basemargin << bitdepth_shift
        clip_high = (1 << bitdepth) - 1 - clip_low
        y = np.clip(y, clip_low, clip_high)
        u = np.clip(u, clip_low, clip_high)
        v = np.clip(v, clip_low, clip_high)

        return y, u, v


class ColorspaceManager:
    def __init__(self):
        # For colorspace parameters, see:
        # - BT.601: https://www.itu.int/rec/R-REC-BT.601
        # - BT.709: https://www.itu.int/rec/R-REC-BT.709
        # - BT.2020: https://www.itu.int/rec/R-REC-BT.2020
        # - BT.2100: https://www.itu.int/rec/R-REC-BT.2100
        self._conversions = {
            ('bt601', 'limited'): Colorspace([0.299, 0.587, 0.114, 1.772, 1.402], (16, 235), (16, 240), 1),
            ('bt601', 'full'): Colorspace([0.299, 0.587, 0.114, 1.772, 1.402], (0, 255), (1, 255), 1),
            ('bt709', 'limited'): Colorspace([0.2126, 0.7152, 0.0722, 1.8556, 1.5748], (16, 235), (16, 240), 1),
            ('bt709', 'full'): Colorspace([0.2126, 0.7152, 0.0722, 1.8556, 1.5748], (0, 255), (1, 255), 1),
            ('bt2020', 'limited'): Colorspace([0.2627, 0.6780, 0.0593, 1.8814, 1.4746], (16, 235), (16, 240), 1),
            ('bt2020', 'full'): Colorspace([0.2627, 0.6780, 0.0593, 1.8814, 1.4746], (0, 255), (1, 255), 1),
            ('bt2100', 'limited'): Colorspace([0.2627, 0.6780, 0.0593, 1.8814, 1.4746], (16, 235), (16, 240), 1),
            ('bt2100', 'full'): Colorspace([0.2627, 0.6780, 0.0593, 1.8814, 1.4746], (0, 255), (1, 255), 1)
        }

    def __getitem__(self, key: Tuple[str, str]):
        identifier, value_range = key
        return self._conversions[(identifier, value_range)]

    def __str__(self):
        return ", ".join(list(map(
            lambda key: f"({key[0]}, {key[1]})",
            self._conversions.keys()
        )))


colorspaces = ColorspaceManager()
